program Example6;

uses
  Windows,
  Messages,
  math,
  dglOpenGL,
  DglShader;

type
  GLMatrix= array[0..15] of GLfloat; // Matrix type
  GLvect= array[0..2] of GLfloat;    // Vector type
  GLBufferf = array of GLfloat;      // Buffer of GLfloat
  GLBufferi = array of GLuint;       // Buffer of GLuint;
  GLBufferb = array of GLubyte;      // Buffer of GLubyte;

const
  WND_TITLE = 'OpenGL Demo';
  FPS_TIMER = 1;                     // Timer to calculate FPS
  FPS_INTERVAL = 1000;               // Calculate FPS every 1000 ms
  OGL3_MAJOR_VERSION_WANTED = 3;     // What OpenGL 3.x is needed
  OGL3_MINOR_VERSION_WANTED = 0;     // What OpenGL 3.x is needed
  OGL3_NEEDED = false;               // Is OpenGL 3.x needed (e.g. the application cannot work without it)

var
  winwidth: integer = 600;           // window x size
  winheight: integer = 400;          // window y size

  h_Wnd  : HWND;                     // Global window handle
  h_DC   : HDC;                      // Global device context
  h_RC   : HGLRC;                    // OpenGL rendering context
  h_RC3  : HGLRC;                    // OpenGL 3.x rendering context
  keys : Array[0..255] of Boolean;   // Holds keystrokes
  FPSCount : Integer = 0;            // Counter for FPS
  ElapsedTime : Integer;             // Elapsed time between frames
  OGL_TITLE: string;     // Info on OpenGL actualy used

  // User vaiables
  xSpeed, ySpeed : glFloat;

  // Mouse Variables
  MouseMove : boolean = false;        // Is the mouse moving?
  MoveXcoord, MoveYcoord : GLfloat;   // Current X and Y position of the cursor

  MouseButton : Integer = -1;         // mouse button down
  Xcoord, Ycoord : Integer;           // Current X and Y position where the mouse was clicked

  OpenGLInitialized : Boolean;        // OpenGL initialize
  OpenGL3Initialized : Boolean;       // OpenGL 3.x initialized

  g_projectionLocation: GLint;        // Location for the ProjectionMatrix

  glslsimplevert: TGLSLShader;        // a Vertex Shader
  glslsimplefrag: TGLSLShader;        // a Fragment Shader
  glslsimpleprog: TGLSLProgram;       // a Shader Program

  g_vao: Gluint;                      // Vertex Array Object for OpenGL 3.x

  g_vertices: GLuint;                 // The VBO for the vertices.
  g_normals: GLuint;                  // The VBO for the normals.
  g_indices: GLuint;                  // The VBO for the indices.

  shapenumindices: integer;           // Number of indices in shape

{$R *.RES}

{------------------------------------------------------------------}
{  Function to convert int to string. (No sysutils = smaller EXE)  }
{------------------------------------------------------------------}
function IntToStr(Num : Integer) : String;  // using SysUtils increase file size by 100K
begin
  Str(Num, result);
end;

{------------------------------------------------------------------}
{  Function to convert float to string. (No sysutils = smaller EXE)}
{------------------------------------------------------------------}
function FloatToStr(Num : single) : String; overload; // using SysUtils increase file size by 100K
begin
  Str(Num, result);
end;

{------------------------------------------------------------------}
{  Function to convert float to string. (No sysutils = smaller EXE)}
{------------------------------------------------------------------}
function FloatToStr(Num : double) : String; overload; // using SysUtils increase file size by 100K
begin
  Str(Num, result);
end;

{------------------------------------------------------------------}
{  Calculate Amount                                                }
{------------------------------------------------------------------}
function amount(vector: GLvect): GLfloat;
begin
  result := sqrt(vector[0] * vector[0] + vector[1] * vector[1] + vector[2] * vector[2]);
end;


{------------------------------------------------------------------}
{  Calculate Normal                                                }
{------------------------------------------------------------------}
procedure normalize(var vector: GLvect);
var
  a: GLfloat;
  i: integer;
begin
	a := amount(vector);

  if (a = 0.0) then
	begin
    	exit;
	end;

	for i := 0 to 2 do
	begin
    	vector[i] := vector[i] / a;
	end;
end;


{------------------------------------------------------------------}
{  Calculate Vector Dot Product                                    }
{------------------------------------------------------------------}
function dotProduct(const vector0: GLvect; const vector1: GLvect): GLfloat;
begin
	result := vector0[0]*vector1[0] + vector0[1]*vector1[1] + vector0[2]*vector1[2];
end;


{------------------------------------------------------------------}
{  Calculate Vector Cross Product                                  }
{------------------------------------------------------------------}
procedure crossProduct(var aresult: GLvect; const vector0: GLvect; const vector1: GLvect);
var
  temp: GLvect;
  i: integer;
begin
	temp[0] := vector0[1]*vector1[2] - vector0[2]*vector1[1];
	temp[1] := vector0[2]*vector1[0] - vector0[0]*vector1[2];
	temp[2] := vector0[0]*vector1[1] - vector0[1]*vector1[0];

	for i := 0 to 2 do
	begin
    	aresult[i] := temp[i];
	end;
end;


{------------------------------------------------------------------}
{  Calculate Identity Matrix (glLoadIdentity)                      }
{------------------------------------------------------------------}
procedure loadIdentity(var matrix: GLmatrix);
begin
	matrix[0] := 1.0;
	matrix[1] := 0.0;
	matrix[2] := 0.0;
	matrix[3] := 0.0;

	matrix[4] := 0.0;
	matrix[5] := 1.0;
	matrix[6] := 0.0;
	matrix[7] := 0.0;

	matrix[8] := 0.0;
	matrix[9] := 0.0;
	matrix[10] := 1.0;
	matrix[11] := 0.0;

	matrix[12] := 0.0;
	matrix[13] := 0.0;
	matrix[14] := 0.0;
	matrix[15] := 1.0;
end;


{------------------------------------------------------------------}
{  Multiply Matrices                                               }
{------------------------------------------------------------------}
procedure multMatrix(var aresult: GLmatrix; const matrix0: GLmatrix; const matrix1: GLmatrix);
var
  i, k: integer;
  temp: GLmatrix;
begin
	for i := 0 to 15 do
	begin
		temp[i] := 0.0;

		for k := 0 to 3 do
		begin
			//			  		row   column   		   row column
			temp[i] := temp[i]+ matrix0[(i mod 4)+(k*4)] * matrix1[k+((i div 4)*4)];
		end;
	end;

	for i := 0 to 15 do
	begin
		aresult[i] := temp[i];
	end;
end;


{------------------------------------------------------------------}
{  Calculate TranslationMatrix (glTranslatef)                      }
{------------------------------------------------------------------}
procedure translateMatrix(var aresult: GLmatrix; x: GLfloat; y: GLfloat; z: GLfloat);
var
  matrix: GLmatrix;
  tempmatrix: GLmatrix;
begin
	loadIdentity(matrix);

	matrix[12] := x;
	matrix[13] := y;
	matrix[14] := z;

  tempmatrix := aresult;
	multMatrix(aresult, tempmatrix, matrix);
end;


{------------------------------------------------------------------}
{  Calculate Rotation Matrix (glrotatef)                           }
{------------------------------------------------------------------}
procedure rotateRzRyRxMatrixf(var aresult: GLmatrix; anglex: GLfloat; angley: GLfloat; anglez: GLfloat);
var
  matrix : GLmatrix;
  tempmatrix : GLmatrix;
  rx : GLfloat;
	ry : GLfloat;
	rz : GLfloat;
	sx : GLfloat;
	cx : GLfloat;
	sy : GLfloat;
	cy : GLfloat;
	sz : GLfloat;
	cz : GLfloat;
begin
	rx := 2.0*PI*anglex/360.0;
	ry := 2.0*PI*angley/360.0;
	rz := 2.0*PI*anglez/360.0;
	sx := sin(rx);
	cx := cos(rx);
	sy := sin(ry);
	cy := cos(ry);
	sz := sin(rz);
	cz := cos(rz);

	loadIdentity(matrix);

	matrix[0] := cy*cz;
	matrix[1] := cy*sz;
	matrix[2] := -sy;

	matrix[4] := cz*sx*sy-cx*sz;
	matrix[5] := cx*cz+sx*sy*sz;
	matrix[6] := cy*sx;

	matrix[8] := cx*cz*sy+sx*sz;
	matrix[9] := -cz*sx+cx*sy*sz;
	matrix[10] := cx*cy;

  tempmatrix := aresult;
	multMatrix(aresult, tempmatrix, matrix);
end;


{------------------------------------------------------------------}
{  Calculate Frustum Matrix (glFrustum)                            }
{------------------------------------------------------------------}
procedure frustum(var aresult: glmatrix; left: GLfloat; right: GLfloat; bottom: GLfloat; top: GLfloat; nearVal: GLfloat; farVal: GLfloat);
begin
	aresult[0] := 2.0*nearVal/(right-left);
	aresult[1] := 0.0;
	aresult[2] := 0.0;
	aresult[3] := 0.0;
	aresult[4] := 0.0;
	aresult[5] := 2.0*nearVal/(top-bottom);
	aresult[6] := 0.0;
	aresult[7] := 0.0;
	aresult[8] := (right+left)/(right-left);
	aresult[9] := (top+bottom)/(top-bottom);
	aresult[10] := -(farVal+nearVal)/(farVal-nearVal);
	aresult[11] := -1.0;
	aresult[12] := 0.0;
	aresult[13] := 0.0;
	aresult[14] := -(2.0*farVal*nearVal)/(farVal-nearVal);
	aresult[15] := 0.0;
end;


{------------------------------------------------------------------}
{  Calculate Perspective Matrix (gluPerspective)                   }
{------------------------------------------------------------------}
procedure perspective(var aresult: glmatrix; fovy: GLfloat; aspect: GLfloat; zNear: GLfloat; zFar: GLfloat);
var
  xmin, xmax, ymin, ymax: GLfloat;
begin
    ymax := zNear * tan(fovy * PI / 360.0);
    ymin := -ymax;
    xmin := ymin * aspect;
    xmax := ymax * aspect;

    frustum(aresult, xmin, xmax, ymin, ymax, zNear, zFar);
end;


{------------------------------------------------------------------}
{  Calculate LookAt Matrix (gluLookat)                             }
{------------------------------------------------------------------}
procedure lookAt(var aresult: GLmatrix; eyeX: GLfloat; eyeY: GLfloat; eyeZ: GLfloat; centerX: GLfloat; centerY: GLfloat; centerZ: GLfloat; upX: GLfloat; upY: GLfloat; upZ: GLfloat );
var
  matrix: GLmatrix;
  tempmatrix : GLmatrix;
  fforward, side, up: GLvect;
begin
  fforward[0] := (centerX - eyeX);
  fforward[1] := (centerY - eyeY);
  fforward[2] := (centerZ - eyeZ);

  normalize(fforward);

  up[0] := upX;
  up[1] := upY;
  up[2] := upZ;

  crossProduct(side, fforward, up);
  normalize(side);

	crossProduct(up, side, fforward);

	matrix[0] := side[0];
	matrix[1] := up[0];
	matrix[2] := -fforward[0];
	matrix[3] := 0.0;
	matrix[4] := side[1];
	matrix[5] := up[1];
	matrix[6] := -fforward[1];
	matrix[7] := 0.0;
	matrix[8] := side[2];
	matrix[9] := up[2];
	matrix[10] := -fforward[2];
	matrix[11] := 0.0;
	matrix[12] := 0.0;
	matrix[13] := 0.0;
	matrix[14] := 0.0;
	matrix[15] := 1.0;

	loadIdentity(tempmatrix);
	multMatrix(aresult, tempmatrix, matrix);

	translateMatrix(aresult, -eyeX, -eyeY, -eyeZ);
end;

//
// Shape sources partly taken from:
//    OpenGL(R) ES 2.0 Programming Guide
//	  http://www.opengles-book.com
//


{------------------------------------------------------------------}
{  Function to create a plane shape                                }
{------------------------------------------------------------------}
function createPlane(halfExtend: GLfloat;var vertices: GLBufferf;var normals: GLbufferf; var texCoords: GLbufferf; var indices: GLBufferi;var numberVertices: GLuint): GluInt;
var
  numVertices: Gluint;
  numIndices: Gluint;
  i: GLuint;

const
  xy_vertices: array[0..15] of GLFloat=
    (		-1.0, -1.0, 0.0, +1.0,
		    +1.0, -1.0, 0.0, +1.0,
		    -1.0, +1.0, 0.0, +1.0,
		    +1.0, +1.0, 0.0, +1.0);
  xy_normals: array[0..11] of GLFloat=
    (		0.0, 0.0, 1.0,
		    0.0, 0.0, 1.0,
		    0.0, 0.0, 1.0,
		    0.0, 0.0, 1.0);
  xy_texCoords: array[0..7] of GLFloat=
    (		0.0, 0.0,
		    1.0, 0.0,
		    0.0, 1.0,
		    1.0, 1.0);
  xy_indices: array[0..5] of GLuint=
    (		0, 1, 2,
		    1, 3, 2);

begin
	numVertices := 4;
	numIndices := 6;

	numberVertices := numVertices;

  SetLength(vertices, 4*numVertices );
  for i := 0 to (4*numVertices) - 1 do
    vertices[i]:=xy_vertices[i];

  for i := 0 to numVertices-1 do
  begin
			vertices[i*4+0] := vertices[i*4+0] * halfExtend;
			vertices[i*4+1] := vertices[i*4+1] * halfExtend;
  end;

  SetLength(normals, 3*numVertices);
  for i := 0 to (3*numVertices) - 1 do
    normals[i]:=xy_normals[i];

  SetLength(texCoords,2*numVertices);
  for i := 0 to (2*numVertices) - 1 do
    texcoords[i]:=xy_texcoords[i];

  SetLength(indices, numIndices);
  for i := 0 to numIndices - 1 do
    indices[i]:=xy_indices[i];

	result := numIndices;
end;


{------------------------------------------------------------------}
{  Function to create a cube shape                                 }
{------------------------------------------------------------------}
function createCube(halfExtend: GLfloat;var vertices: GLBufferf;var normals: GLbufferf; var texCoords: GLbufferf; var indices: GLBufferi;var numberVertices: GLuint): GLuint;
var
  numVertices : GLuint;
	numIndices : GLuint;
  i: GLuint;
const
  cubeVertices: array[0..95] of GLFloat=
	(
	  -1.0, -1.0, -1.0, +1.0,
	  -1.0, -1.0, +1.0, +1.0,
	  +1.0, -1.0, +1.0, +1.0,
	  +1.0, -1.0, -1.0, +1.0,
	  -1.0, +1.0, -1.0, +1.0,
	  -1.0, +1.0, +1.0, +1.0,
	  +1.0, +1.0, +1.0, +1.0,
	  +1.0, +1.0, -1.0, +1.0,
	  -1.0, -1.0, -1.0, +1.0,
	  -1.0, +1.0, -1.0, +1.0,
	  +1.0, +1.0, -1.0, +1.0,
	  +1.0, -1.0, -1.0, +1.0,
	  -1.0, -1.0, +1.0, +1.0,
	  -1.0, +1.0, +1.0, +1.0,
	  +1.0, +1.0, +1.0, +1.0,
	  +1.0, -1.0, +1.0, +1.0,
	  -1.0, -1.0, -1.0, +1.0,
	  -1.0, -1.0, +1.0, +1.0,
	  -1.0, +1.0, +1.0, +1.0,
	  -1.0, +1.0, -1.0, +1.0,
	  +1.0, -1.0, -1.0, +1.0,
	  +1.0, -1.0, +1.0, +1.0,
	  +1.0, +1.0, +1.0, +1.0,
	  +1.0, +1.0, -1.0, +1.0
	);

	cubeNormals: array[0..71] of GLFloat=
	(
	   0.0, -1.0,  0.0,
	   0.0, -1.0,  0.0,
	   0.0, -1.0,  0.0,
	   0.0, -1.0,  0.0,
	   0.0, +1.0,  0.0,
	   0.0, +1.0,  0.0,
	   0.0, +1.0,  0.0,
	   0.0, +1.0,  0.0,
	   0.0,  0.0, -1.0,
	   0.0,  0.0, -1.0,
	   0.0,  0.0, -1.0,
	   0.0,  0.0, -1.0,
	   0.0,  0.0, +1.0,
	   0.0,  0.0, +1.0,
	   0.0,  0.0, +1.0,
	   0.0,  0.0, +1.0,
	  -1.0,  0.0,  0.0,
	  -1.0,  0.0,  0.0,
	  -1.0,  0.0,  0.0,
	  -1.0,  0.0,  0.0,
	  +1.0,  0.0,  0.0,
	  +1.0,  0.0,  0.0,
	  +1.0,  0.0,  0.0,
	  +1.0,  0.0,  0.0
	);
	cubeTexCoords: array[0..47] of GLFloat=
	(
	  0.0, 0.0,
	  0.0, 1.0,
	  1.0, 1.0,
	  1.0, 0.0,
	  1.0, 0.0,
	  1.0, 1.0,
	  0.0, 1.0,
	  0.0, 0.0,
	  0.0, 0.0,
	  0.0, 1.0,
	  1.0, 1.0,
	  1.0, 0.0,
	  0.0, 0.0,
	  0.0, 1.0,
	  1.0, 1.0,
	  1.0, 0.0,
	  0.0, 0.0,
	  0.0, 1.0,
	  1.0, 1.0,
	  1.0, 0.0,
	  0.0, 0.0,
	  0.0, 1.0,
	  1.0, 1.0,
	  1.0, 0.0
	);

	cubeIndices: array[0..35] of GLuint=
	(
	 0, 2, 1,
	 0, 3, 2,
	 4, 5, 6,
	 4, 6, 7,
	 8, 9, 10,
	 8, 10, 11,
	 12, 15, 14,
	 12, 14, 13,
	 16, 17, 18,
	 16, 18, 19,
	 20, 23, 22,
	 20, 22, 21
	);

begin
	numVertices := 24;
	numIndices := 36;

	numberVertices := numVertices;


  SetLength(vertices, 4*numVertices );
  for i := 0 to (4*numVertices) - 1 do
    vertices[i]:=cubevertices[i];

  for i := 0 to numVertices-1 do
  begin
    vertices[i*4+0] := vertices[i*4+0] * halfExtend;
    vertices[i*4+1] := vertices[i*4+1] * halfExtend;
    vertices[i*4+2] := vertices[i*4+2] * halfExtend;
  end;

  SetLength(normals, 3*numVertices);
  for i := 0 to (3*numVertices) - 1 do
    normals[i]:=cubenormals[i];

  SetLength(texCoords,2*numVertices);
  for i := 0 to (2*numVertices) - 1 do
    texcoords[i]:=cubetexcoords[i];

  SetLength(indices, numIndices);
  for i := 0 to numIndices - 1 do
    indices[i]:=cubeindices[i];

	result := numIndices;
end;


{------------------------------------------------------------------}
{  Function to create a sphere shape                               }
{------------------------------------------------------------------}
function createSphere(numSlices:GLuint; radius: GLfloat; var vertices: GLBufferf;var normals: GLbufferf; var texCoords: GLbufferf; var indices: GLBufferi;var numberVertices: GLuint): GLuint;
var
  numParallels : GLuint;
	numVertices : GLuint;
	numIndices : GLuint;
  angleStep : GLfloat;
  i,j,k: GLuint;
  vertexIndex: GLuint;
  normalIndex: Gluint;
  texCoordsIndex: GLuint;
begin
	numParallels := numSlices;
	numVertices := ( numParallels + 1 ) * ( numSlices + 1 );
	numIndices := numParallels * numSlices * 6;

	angleStep := (2.0 * PI) / numSlices;

	numberVertices := numVertices;

  SetLength(vertices, 4*numVertices);

  SetLength(normals, 3*numVertices);

  SetLength(texCoords, 2*numVertices);

  SetLength(indices, numIndices);

	for i := 0 to numParallels do
	begin
		for j := 0 to numSlices do
		begin
      vertexIndex := ( i * (numSlices + 1) + j ) * 4;
      vertices[vertexIndex + 0] := radius * sin ( angleStep * i ) * sin ( angleStep * j );
      vertices[vertexIndex + 1] := radius * cos ( angleStep * i );
      vertices[vertexIndex + 2] := radius * sin ( angleStep * i ) * cos ( angleStep * j );
      vertices[vertexIndex + 3] := 1.0;

      vertexIndex := ( i * (numSlices + 1) + j ) * 4;
      normalIndex := ( i * (numSlices + 1) + j ) * 3;
      normals[normalIndex + 0] := vertices[vertexIndex + 0] / radius;
      normals[normalIndex + 1] := vertices[vertexIndex + 1] / radius;
      normals[normalIndex + 2] := vertices[vertexIndex + 2] / radius;

      texCoordsIndex := ( i * (numSlices + 1) + j ) * 2;
      texCoords[texCoordsIndex + 0] := j / numSlices;
      texCoords[texCoordsIndex + 1] := ( 1.0 - i ) / (numParallels - 1 );
		end;
	end;

  k:=0;
  for i := 0 to numParallels-1 do
    for j := 0 to numSlices-1 do
    begin
      indices[k+0] := i * ( numSlices + 1 ) + j;
      indices[k+1] := ( i + 1 ) * ( numSlices + 1 ) + j;
      indices[k+2] := ( i + 1 ) * ( numSlices + 1 ) + ( j + 1 );

      indices[k+3] := i * ( numSlices + 1 ) + j;
      indices[k+4] := ( i + 1 ) * ( numSlices + 1 ) + ( j + 1 );
      indices[k+5] := i * ( numSlices + 1 ) + ( j + 1 );
      k:=k+6;
    end;

	result := numIndices;
end;


{------------------------------------------------------------------}
{  Function to swap color channes in bitmap data                   }
{------------------------------------------------------------------}
procedure swapColorChannel(width: GLuint; height: GLuint; format: GLenum;var pixelData: GLBufferb);
var
  temp: GLubyte;
  bytesPerPixel: GLubyte;
  i: GLuint;
begin
	bytesPerPixel := 3;

	if (format = GL_RGBA) then
	  bytesPerPixel := 4;

	// swap the R and B values to get RGB since the bitmap color format is in BGR
  i:=0;
	while i <(width*height*bytesPerPixel)-1 do
	begin
		temp := pixelData[i];
		pixelData[i] := pixelData[i+2];
		pixelData[i+2] := temp;
    i:=i+bytesPerPixel;
	end;
end;


{------------------------------------------------------------------}
{  Function to load .tga file                                      }
{------------------------------------------------------------------}
function loadTGA(const filename: pansichar;var width: GLushort; var height: GLushort; var format: GLenum; var pixelData: GLbufferb): GLboolean;
var
  lfile: file;
	ltype: Glubyte;
  buf: GluByte;
	bitsPerPixel: GLubyte;
  sizeread: integer;
  pixelsRead: GLuint;
  amount: GLubyte;
  i,j,k: Gluint;

  ColorDepth    : Integer;
  ImageSize     : Integer;
  BufferIndex : Integer;
  currentByte : Integer;
  currentPixel : Integer;
    //Front: Byte;
  comppixeldata: GLbufferb;
begin
	// check, if we have a valid pointer
	if (filename='') then
  begin
		result := false;
    exit;
	end;

	// open filename in "read binary" mode
	AssignFile(lfile, filename);
  Reset( lfile,1 ); //Read bytes

 	// seek through the tga header, up to the type:
  BlockRead(lfile,buf,1);
  BlockRead(lfile,buf,1);

 	// read the type
  BlockRead(lfile,ltype,1);

	// check the type
	if (ltype <> 2) and (ltype <> 10) and (ltype <> 11) then
	begin
		close(lfile);
		result := false;
    exit;
	end;

 	// seek through the tga header, up to the width/height:
  BlockRead(lfile,buf,1);
  BlockRead(lfile,buf,1);
  BlockRead(lfile,buf,1);
  BlockRead(lfile,buf,1);
  BlockRead(lfile,buf,1);
  BlockRead(lfile,buf,1);
  BlockRead(lfile,buf,1);
  BlockRead(lfile,buf,1);
  BlockRead(lfile,buf,1);

 	// read the width
  BlockRead(lfile,width,2,sizeread);

 	// read the height
  BlockRead(lfile,height,2,sizeread);

	// read the bits per pixel
  BlockRead(lfile,bitsPerPixel,1,sizeread);

	// check the pixel depth
	if (bitsPerPixel <> 8) and (bitsPerPixel <> 24) and (bitsPerPixel <> 32) then
  begin
		close(lfile);
		result := false;
    exit;
	end
	else
	begin
		format := GL_ALPHA;
		if (bitsPerPixel = 24) then
		  format := GL_RGB
		else if (bitsPerPixel = 32) then
			format := GL_RGBA;
	end;

	// move file pointer to beginning of targa  data
  BlockRead(lfile,buf,1);

	// allocate enough memory for the targa  data
  SetLength(pixeldata,width*height*bitsPerPixel div 8);


	if (ltype = 2) then
	begin
		// read in the raw data
		BlockRead(lfile,pixelData[0], width*height*bitsPerPixel div 8, sizeread);
	end
	else
	begin
		// RLE encoded
    ColorDepth :=bitsperpixel DIV 8;
    CurrentByte :=0;
    CurrentPixel :=0;
    BufferIndex :=0;

    //Read Compressed Data
    SetLength(comppixeldata,FileSize(lfile)-18);
    BlockRead(lfile,comppixelData[0], FileSize(lfile)-18, sizeread);

    repeat
        Amount := comppixeldata[BufferIndex];
        Inc(BufferIndex);
        if Amount < 128 then
        begin
          For I := 0 to Amount do
          begin
            for J := 0 to ColorDepth - 1 do
              pixeldata[CurrentByte+J] := comppixeldata[BufferIndex+i*ColorDepth+J];
            CurrentByte := CurrentByte + ColorDepth;
            inc(CurrentPixel);
          end;
          BufferIndex :=BufferIndex + (Amount+1)*ColorDepth
        end
        else
        begin
          For I := 0 to Amount -128 do
          begin
            for J := 0 to ColorDepth - 1 do
              pixeldata[CurrentByte+J] := comppixeldata[BufferIndex+J];
            CurrentByte := CurrentByte + ColorDepth;
            inc(CurrentPixel);
          end;
          BufferIndex :=BufferIndex + ColorDepth
        end;
      until CurrentPixel >= Width*Height;

      SetLength(comppixeldata,0); //Clean Up
	end;

	// swap the color if necessary
	if (bitsPerPixel = 24) or (bitsPerPixel = 32) then
	begin
		swapColorChannel(width, height, format, pixelData);
	end;

	// close the file
	close(lfile);

	result := true;
end;


{------------------------------------------------------------------}
{  Function to load .bmp file                                      }
{------------------------------------------------------------------}
function loadBMP(const filename: pansichar;var width: GLushort; var height: GLushort; var format: GLenum; var pixelData: GLbufferb): GLboolean;
var
  lfile: file;
  FileHeader: TBITMAPFILEHEADER;
  InfoHeader: TBITMAPINFOHEADER;
  Palette: array of RGBQUAD;
  sizeread: integer;
  BitmapLength: LongWord;
  PaletteLength: LongWord;
begin
	// check, if we have a valid pointer
	if (filename='') then
  begin
		result := false;
    exit;
	end;

	// open filename in "read binary" mode
	AssignFile(lfile, filename);
  Reset( lfile,1 ); //Read Bytes

  //Read Headers ...
  BlockRead(lfile, FileHeader, SizeOf(FileHeader), sizeread);
  BlockRead(lfile, InfoHeader, SizeOf(InfoHeader), sizeread);

  // Get palette
  PaletteLength := InfoHeader.biClrUsed;
  SetLength(Palette, PaletteLength);
  BlockRead(lfile, Palette, PaletteLength, sizeread);
  if (sizeread <> PaletteLength) then begin
    result := false;
    Exit;
  end;

  Width := InfoHeader.biWidth;
  Height := InfoHeader.biHeight;
  BitmapLength := InfoHeader.biSizeImage;
  if BitmapLength = 0 then
    BitmapLength := Width * Height * InfoHeader.biBitCount Div 8;

  // check the pixel depth
	if (InfoHeader.biBitCount <> 8) and (InfoHeader.biBitCount <> 24) and (InfoHeader.biBitCount <> 32) then
  begin
		close(lfile);
		result := false;
    exit;
	end
	else
	begin
		format := GL_ALPHA;
		if (InfoHeader.biBitCount = 24) then
		  format := GL_RGB
		else if (InfoHeader.biBitCount = 32) then
			format := GL_RGBA;
	end;

  // Get the actual pixel data
  SetLength(pixeldata, BitmapLength);
  BlockRead(lfile, pixeldata[0], BitmapLength, sizeread);
  if (sizeread <> BitmapLength) then begin
    result := false;
    Exit;
  end;

  // swap the color if necessary
	if (InfoHeader.biBitCount = 24) or (InfoHeader.biBitCount = 32) then
	begin
		swapColorChannel(width, height, format, pixelData);
	end;

  Close(lfile);

  result := true;
end;

{------------------------------------------------------------------}
{  Function to draw the actual scene                               }
{------------------------------------------------------------------}
procedure glDraw();
var
	g_modelViewLocation: GLint;
	model: GLmatrix;
  tempmatrix: GLmatrix;
  g_modelView: GLmatrix;
begin
  glClear(GL_COLOR_BUFFER_BIT or GL_DEPTH_BUFFER_BIT);

  // Calculate the model matrix ...
  loadIdentity(model);
	rotateRzRyRxMatrixf(model, 30.0*xspeed, 30.0*yspeed, 0.0); //cube can now be move by arrow keys ...
	// ... and the view matrix ...
	lookAt(tempmatrix, 0.0, 0.0, 5.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);
	// ... to get the final model view matrix
	multMatrix(g_modelView, tempmatrix, model);

  // http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml
	glUniformMatrix4fv(g_modelViewLocation, 1, false, @g_modelView);

  if OpenGL3Initialized then
  begin
    glBindVertexArray(g_vao); //Tell what vao we want to draw
  end;

	// http://www.opengl.org/sdk/docs/man/xhtml/glDrawElements.xml
	glDrawElements(GL_TRIANGLES, shapenumindices, GL_UNSIGNED_INT, nil);

  if OpenGL3Initialized then
  begin
    glBindVertexArray(0); //No vao to use
  end;
end;


{------------------------------------------------------------------}
{  Initialise OpenGL                                               }
{------------------------------------------------------------------}
procedure glInit();
var
	g_modelViewLocation: GLint;
  g_textureLocation: GLint;
	g_vertexLocation: GLint;
	g_normalLocation: GLint;
  g_texCoordLocation: GLint;
  g_texCoords: GLint;
  g_texture: GLint;

  //Move the cube along the minus z axis for making it visible. Also rotated.
  g_modelView: GLmatrix;
  // Matrix for the model
	model: GLmatrix;
  tempmatrix: GLmatrix;

  shapenumvertices: GLuint;

  shapepoints: GLbufferf;
	shapenormals: GLbufferf;
	shapeindices: GLbufferi;
  shapetexcoords: GLBufferf;

  width: GLushort;
  height: GLushort;
  format: GLenum;
  pixelData: GLbufferb;

begin
  //create a shape
  //shapenumindices := createPlane(1.0,shapepoints,shapenormals,shapetexcoords,shapeindices,shapenumvertices);
  shapenumindices := createCube(1.0,shapepoints,shapenormals,shapetexcoords,shapeindices,shapenumvertices);
  //shapenumIndices := createSphere(32, 1.0, shapepoints, shapenormals, shapetexcoords, shapeindices, shapenumVertices);

  //load simple glsl shader
  glslsimpleprog := TGLSLProgram.Create();
  glslsimplevert := TGLSLShader.Create('shaders\Vertex.vs');
  glslsimplefrag := TGLSLShader.Create('shaders\Fragment.fs', GL_FRAGMENT_SHADER_ARB);

  glslsimpleprog.Attach(glslsimplevert);
  glslsimpleprog.Attach(glslsimplefrag);
  glslsimpleprog.Link;
  glslsimpleprog.Enable; //Enable once is enough if only one glsl prog is used...

  // http://www.opengl.org/sdk/docs/man/xhtml/glGetUniformLocation.xml
	g_projectionLocation := glslsimpleprog.GetUniformLocation('projectionMatrix');

	// http://www.opengl.org/sdk/docs/man/xhtml/glGetUniformLocation.xml
	g_modelViewLocation := glslsimpleprog.GetUniformLocation('modelViewMatrix');

  // http://www.opengl.org/sdk/docs/man/xhtml/glGetUniformLocation.xml
	g_textureLocation := glslsimpleprog.GetUniformLocation('firstTexture');

	// http://www.opengl.org/sdk/docs/man/xhtml/glGetAttribLocation.xml
	g_vertexLocation := glslsimpleprog.GetAttribLocation('vertex');

	// http://www.opengl.org/sdk/docs/man/xhtml/glGetAttribLocation.xml
	g_normalLocation := glslsimpleprog.GetAttribLocation('normal');

  // http://www.opengl.org/sdk/docs/man/xhtml/glGetAttribLocation.xml
	g_texCoordLocation := glslsimpleprog.GetAttribLocation('texCoord');

  if OpenGL3Initialized then
  begin
    // One VAO allocation
    glGenVertexArrays(1,@g_vao);

    // First VAO setup
    glBindVertexArray(g_vao);
  end;

  //Vertices buffer
  glGenBuffers(1, @g_vertices);
  glBindBuffer(GL_ARRAY_BUFFER, g_vertices);
  glBufferData(GL_ARRAY_BUFFER, shapenumvertices*4*sizeof(GLfloat), shapepoints, GL_STATIC_DRAW);
	glVertexAttribPointer(g_vertexLocation, 4, GL_FLOAT, false, 0, 0);
	glEnableVertexAttribArray(g_vertexLocation);
  setLength(shapepoints,0);

  //Normals buffer
  glGenBuffers(1, @g_normals);
  glBindBuffer(GL_ARRAY_BUFFER, g_normals);
  glBufferData(GL_ARRAY_BUFFER,shapenumvertices*3*sizeof(GLfloat), shapenormals, GL_STATIC_DRAW);
	glVertexAttribPointer(g_normalLocation, 3, GL_FLOAT, false, 0, 0);
	glEnableVertexAttribArray(g_normalLocation);
  setLength(shapenormals,0);

  //TexCoords buffer
  // http://www.opengl.org/sdk/docs/man/xhtml/glGenBuffers.xml
	glGenBuffers(1, @g_texCoords);
	// http://www.opengl.org/sdk/docs/man/xhtml/glBindBuffer.xml
	glBindBuffer(GL_ARRAY_BUFFER, g_texCoords);
	// http://www.opengl.org/sdk/docs/man/xhtml/glBufferData.xml
	glBufferData(GL_ARRAY_BUFFER, shapenumVertices*2*sizeof(GLfloat), shapetexCoords, GL_STATIC_DRAW);
	glVertexAttribPointer(g_texcoordLocation, 2, GL_FLOAT, false, 0, 0);
	glEnableVertexAttribArray(g_texcoordLocation);
  setLength(shapetexcoords,0);

  //Indices buffer
  glGenBuffers(1, @g_indices);
	glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, g_indices);
	glBufferData(GL_ELEMENT_ARRAY_BUFFER, shapenumindices*sizeof(GLuint), shapeindices, GL_STATIC_DRAW);
  setLength(shapeindices,0);

  //Load Texture
  loadTGA('Textures\crate.tga', width, height, format, pixelData);

	// http://www.opengl.org/documentation/specs/man_pages/hardcopy/GL/html/gl/gentextures.html
	glGenTextures(1, @g_texture);

	// http://www.opengl.org/documentation/specs/man_pages/hardcopy/GL/html/gl/bindtexture.html
	glBindTexture(GL_TEXTURE_2D, g_texture);

	// http://www.opengl.org/documentation/specs/man_pages/hardcopy/GL/html/gl/teximage2d.html
	glTexImage2D(GL_TEXTURE_2D, 0, format, width, height, 0, format, GL_UNSIGNED_BYTE, pixelData);
	//delete[] pixelData;
  SetLength(pixelData,0);

	// http://www.opengl.org/documentation/specs/man_pages/hardcopy/GL/html/gl/texparameter.html
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	// http://www.opengl.org/documentation/specs/man_pages/hardcopy/GL/html/gl/texparameter.html
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);

	// http://www.opengl.org/documentation/specs/man_pages/hardcopy/GL/html/gl/texparameter.html
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);

	// http://www.opengl.org/documentation/specs/man_pages/hardcopy/GL/html/gl/texparameter.html
	glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);

  // http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml
	glUniform1i(g_textureLocation, 0);

  // Calculate the model matrix ...
  loadIdentity(model);
	rotateRzRyRxMatrixf(model, 30.0, 30.0, 0.0);
	// ... and the view matrix ...
	lookAt(tempmatrix, 0.0, 0.0, 5.0, 0.0, 0.0, 0.0, 0.0, 1.0, 0.0);
	// ... to get the final model view matrix
	multMatrix(g_modelView, tempmatrix, model);

  // http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml
	glUniformMatrix4fv(g_modelViewLocation, 1, false, @g_modelView);

  if OpenGL3Initialized then
  begin
    glBindVertexArray(0); //Done creating the vao
  end;

	// http://www.opengl.org/documentation/specs/man_pages/hardcopy/GL/html/gl/clearcolor.html
	glClearColor(0.0, 0.0, 0.0, 0.0);

	// http://www.opengl.org/documentation/specs/man_pages/hardcopy/GL/html/gl/cleardepth.html
	glClearDepth(1.0);

	//http://www.opengl.org/documentation/specs/man_pages/hardcopy/GL/html/gl/enable.html
	glEnable(GL_DEPTH_TEST);

	//http://www.opengl.org/documentation/specs/man_pages/hardcopy/GL/html/gl/enable.html
	glEnable(GL_CULL_FACE);
end;


{------------------------------------------------------------------}
{  De-Initialize OpenGL                                            }
{------------------------------------------------------------------}
procedure glExit();
begin
  if OpenGL3Initialized then
  begin
    // http://www.opengl.org/sdk/docs/man/xhtml/glDeleteVertexArrays.xml
    glDeleteVertexArrays(1, @g_vao);
  end;

  // http://www.opengl.org/sdk/docs/man/xhtml/glDeleteBuffers.xml
  glDeleteBuffers(1, @g_vertices);
  g_vertices := 0;
	glDeleteBuffers(1, @g_normals);
  g_normals := 0;
	glDeleteBuffers(1, @g_indices);
  g_indices := 0;

  glslsimpleprog.Free();
  glslsimplefrag.Free();
  glslsimplevert.Free();
end;


{------------------------------------------------------------------}
{  Handle window resize                                            }
{------------------------------------------------------------------}
procedure glResizeWnd(Width, Height : Integer);
var
  //Projection matrix will be calculated later.
  g_projection: GLMatrix;
begin
  if (Height=0) then
     Height:=1;
  glViewport(0, 0, Width, Height);

  // Calculate the projection matrix and set it
	perspective(g_projection, 40.0, Width/Height, 1.0, 100.0);

	// http://www.opengl.org/sdk/docs/man/xhtml/glUniform.xml
	glUniformMatrix4fv(g_projectionLocation, 1, false, @g_projection);
end;


{------------------------------------------------------------------}
{  Processes all the mouse move                                  }
{------------------------------------------------------------------}
procedure ProcessMouseMove();
begin
  if MouseMove then
  begin
    // Add Code here to execute on mouse move
    MouseMove := False;
  end;
end;


{------------------------------------------------------------------}
{  Processes all the mouse clicks                                  }
{------------------------------------------------------------------}
procedure ProcessMouseClick;
begin
  case MouseButton of
  1: // Left Mouse Button
    begin
      // Add code to execute on mouse click
      MouseButton := 0; // Cancel our mouse click (To use this procedure as a mouse down event remove this line)
    end;
  2: // Right Mouse Button
    begin
      // Add code to execute on mouse click
      MouseButton := 0; // Cancel our mouse click (To use this procedure as a mouse down event remove this line)
    end;
  3: // Middle Mouse Button
    begin
      // Add code to execute on mouse click
      MouseButton := 0;  // Cancel our mouse click (To use this procedure as a mouse down event remove this line)
    end;
  end;
end;


{------------------------------------------------------------------}
{  Processes all the keystrokes                                    }
{------------------------------------------------------------------}
procedure ProcessKeys;
begin
  if (keys[VK_UP])    then xspeed := xspeed - 0.002;
  if (keys[VK_DOWN])  then xspeed := xspeed + 0.002;
  if (keys[VK_RIGHT]) then yspeed := yspeed + 0.002;
  if (keys[VK_LEFT])  then yspeed := yspeed - 0.002;
end;


{------------------------------------------------------------------}
{  Determines the applications response to the messages received  }
{------------------------------------------------------------------}
function WndProc(hWnd: HWND; Msg: UINT;  wParam: WPARAM;  lParam: LPARAM): LRESULT; stdcall;
begin
  case (Msg) of
    WM_CREATE:
      begin
        // Insert stuff you want executed when the program starts
      end;
    WM_CLOSE:
      begin
        PostQuitMessage(0);
        Result := 0
      end;
    WM_KEYDOWN:       // Set the pressed key (wparam) to equal true so we can check if its pressed
      begin
        keys[wParam] := True;
        Result := 0;
      end;
    WM_KEYUP:         // Set the released key (wparam) to equal false so we can check if its pressed
      begin
        keys[wParam] := False;
        Result := 0;
      end;
    WM_MOUSEMOVE:
      begin
        MouseMove := True;
        MoveXcoord := LOWORD(lParam);
        MoveYcoord := HIWORD(lParam);
        Result := 0;
      end;
    WM_LBUTTONDOWN:
      begin
        ReleaseCapture();   // need them here, because if mouse moves off
        SetCapture(h_Wnd);  // window and returns, it needs to reset status
        MouseButton := 1;
        Xcoord := LOWORD(lParam);
        Ycoord := HIWORD(lParam);
        Result := 0;
      end;
    WM_RBUTTONDOWN:
      begin
        ReleaseCapture();   // need them here, because if mouse moves off
        SetCapture(h_Wnd);  // window and returns, it needs to reset status
        MouseButton := 2;
        Xcoord := LOWORD(lParam);
        Ycoord := HIWORD(lParam);
        Result := 0;
      end;
    WM_MBUTTONDOWN:
      begin
        ReleaseCapture();   // need them here, because if mouse moves off
        SetCapture(h_Wnd);  // window and returns, it needs to reset status
        MouseButton := 3;
        Xcoord := LOWORD(lParam);
        Ycoord := HIWORD(lParam);
        Result := 0;
      end;
    WM_LBUTTONUP,WM_RBUTTONUP,WM_MBUTTONUP:
      begin
        ReleaseCapture();   // above
        MouseButton := 0;
        XCoord := 0;
        YCoord := 0;
        Result := 0;
      end;
    WM_SIZE:          // Resize the window with the new width and height
      begin
        if OpenGLInitialized then
         glResizeWnd(LOWORD(lParam),HIWORD(lParam));
        Result := 0;
      end;
    WM_TIMER :                     // Add code here for all timers to be used.
      begin
        if wParam = FPS_TIMER then
        begin
          FPSCount :=Round(FPSCount * 1000/FPS_INTERVAL);   // calculate to get per Second incase intercal is less or greater than 1 second
          SetWindowText(h_Wnd, PChar(WND_TITLE + ' ('+OGL_TITLE+') [' + intToStr(FPSCount) + ' FPS]'));
          FPSCount := 0;
          Result := 0;
        end;
      end;
    else
      begin

      Result := DefWindowProc(hWnd, Msg, wParam, lParam);    // Default result if nothing happens
      end;
  end;
end;


{---------------------------------------------------------------------}
{  Properly destroys the window created at startup (no memory leaks)  }
{---------------------------------------------------------------------}
procedure glKillWnd(Fullscreen : Boolean);
begin
  if Fullscreen then             // Change back to non fullscreen
  begin
    ChangeDisplaySettings(devmode(nil^), 0);
    ShowCursor(True);
  end;

  //Call glExit to remove user set OpenGL things
  glExit();

  // Makes current rendering context not current, and releases the device
  // context that is used by the rendering context.
  if (not wglMakeCurrent(h_DC, 0)) then
    MessageBox(0, 'Release of DC and RC failed!', 'Error', MB_OK or MB_ICONERROR);

  // Attempts to delete the rendering context normal or OpenGL 3.x
  if OpenGL3Initialized then
  begin
    if (not wglDeleteContext(h_RC3)) then
    begin
      MessageBox(0, 'Release of rendering context failed!', 'Error', MB_OK or MB_ICONERROR);
      h_RC3 := 0;
    end;
  end else
  begin
    if (not wglDeleteContext(h_RC)) then
    begin
      MessageBox(0, 'Release of rendering context failed!', 'Error', MB_OK or MB_ICONERROR);
      h_RC := 0;
    end;
  end;

  // Attemps to release the device context
  if ((h_DC > 0) and (ReleaseDC(h_Wnd, h_DC) = 0)) then
  begin
    MessageBox(0, 'Release of device context failed!', 'Error', MB_OK or MB_ICONERROR);
    h_DC := 0;
  end;

  // Attempts to destroy the window
  if ((h_Wnd <> 0) and (not DestroyWindow(h_Wnd))) then
  begin
    MessageBox(0, 'Unable to destroy window!', 'Error', MB_OK or MB_ICONERROR);
    h_Wnd := 0;
  end;

  // Attempts to unregister the window class
  if (not UnRegisterClass('DGL_OpenGL', hInstance)) then
  begin
    MessageBox(0, 'Unable to unregister window class!', 'Error', MB_OK or MB_ICONERROR);
    hInstance := 0;
  end;
end;


{--------------------------------------------------------------------}
{  Creates the window and attaches a OpenGL rendering context to it  }
{--------------------------------------------------------------------}
function glCreateWnd(Width, Height : Integer; Fullscreen : Boolean; PixelDepth : Integer) : Boolean;
var
  wndClass : TWndClass;         // Window class
  dwStyle : DWORD;              // Window styles
  dwExStyle : DWORD;            // Extended window styles
  dmScreenSettings : DEVMODE;   // Screen settings (fullscreen, etc...)
  PixelFormat : GLuint;         // Settings for the OpenGL rendering
  h_Instance : HINST;           // Current instance
  pfd : TPIXELFORMATDESCRIPTOR;  // Settings for the OpenGL window
  attribs: array[0..6] of integer; //OpenGL 3.x Attribs
begin
  InitOpenGL;                               // New call to initialize and bind the OpenGL dll

  h_Instance := GetModuleHandle(nil);       //Grab An Instance For Our Window
  ZeroMemory(@wndClass, SizeOf(wndClass));  // Clear the window class structure

  with wndClass do                    // Set up the window class
  begin
    style         := CS_HREDRAW or    // Redraws entire window if length changes
                     CS_VREDRAW or    // Redraws entire window if height changes
                     CS_OWNDC;        // Unique device context for the window
    lpfnWndProc   := @WndProc;        // Set the window procedure to our func WndProc
    hInstance     := h_Instance;
    hCursor       := LoadCursor(0, IDC_ARROW);
    lpszClassName := 'DGL_OpenGL';
  end;

  if (RegisterClass(wndClass) = 0) then  // Attemp to register the window class
  begin
    MessageBox(0, 'Failed to register the window class!', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit
  end;

  // Change to fullscreen if so desired
  if Fullscreen then
  begin
    ZeroMemory(@dmScreenSettings, SizeOf(dmScreenSettings));
    with dmScreenSettings do begin              // Set parameters for the screen setting
      dmSize       := SizeOf(dmScreenSettings);
      dmPelsWidth  := Width;                    // Window width
      dmPelsHeight := Height;                   // Window height
      dmBitsPerPel := PixelDepth;               // Window color depth
      dmFields     := DM_PELSWIDTH or DM_PELSHEIGHT or DM_BITSPERPEL;
    end;

    // Try to change screen mode to fullscreen
    if (ChangeDisplaySettings(dmScreenSettings, CDS_FULLSCREEN) = DISP_CHANGE_FAILED) then
    begin
      MessageBox(0, 'Unable to switch to fullscreen!', 'Error', MB_OK or MB_ICONERROR);
      Fullscreen := False;
    end;
  end;

  // If we are still in fullscreen then
  if (Fullscreen) then
  begin
    dwStyle := WS_POPUP or                // Creates a popup window
               WS_CLIPCHILDREN            // Doesn't draw within child windows
               or WS_CLIPSIBLINGS;        // Doesn't draw within sibling windows
    dwExStyle := WS_EX_APPWINDOW;         // Top level window
    ShowCursor(False);                    // Turn of the cursor (gets in the way)
  end
  else
  begin
    dwStyle := WS_OVERLAPPEDWINDOW or     // Creates an overlapping window
               WS_CLIPCHILDREN or         // Doesn't draw within child windows
               WS_CLIPSIBLINGS;           // Doesn't draw within sibling windows
    dwExStyle := WS_EX_APPWINDOW or       // Top level window
                 WS_EX_WINDOWEDGE;        // Border with a raised edge
  end;

  // Attempt to create the actual window
  h_Wnd := CreateWindowEx(dwExStyle,      // Extended window styles
                          'DGL_OpenGL',       // Class name
                          WND_TITLE,      // Window title (caption)
                          dwStyle,        // Window styles
                          0, 0,           // Window position
                          Width, Height,  // Size of window
                          0,              // No parent window
                          0,              // No menu
                          h_Instance,     // Instance
                          nil);           // Pass nothing to WM_CREATE

  if h_Wnd = 0 then
  begin
    glKillWnd(Fullscreen);                // Undo all the settings we've changed
    MessageBox(0, 'Unable to create window!', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;

  // Try to get a device context
  h_DC := GetDC(h_Wnd);
  if (h_DC = 0) then
  begin
    glKillWnd(Fullscreen);
    MessageBox(0, 'Unable to get a device context!', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;

  // Settings for the OpenGL window
  with pfd do
  begin
    nSize           := SizeOf(TPIXELFORMATDESCRIPTOR); // Size Of This Pixel Format Descriptor
    nVersion        := 1;                    // The version of this data structure
    dwFlags         := PFD_DRAW_TO_WINDOW    // Buffer supports drawing to window
                       or PFD_SUPPORT_OPENGL // Buffer supports OpenGL drawing
                       or PFD_DOUBLEBUFFER;  // Supports double buffering
    iPixelType      := PFD_TYPE_RGBA;        // RGBA color format
    cColorBits      := PixelDepth;           // OpenGL color depth
    cRedBits        := 0;                    // Number of red bitplanes
    cRedShift       := 0;                    // Shift count for red bitplanes
    cGreenBits      := 0;                    // Number of green bitplanes
    cGreenShift     := 0;                    // Shift count for green bitplanes
    cBlueBits       := 0;                    // Number of blue bitplanes
    cBlueShift      := 0;                    // Shift count for blue bitplanes
    cAlphaBits      := 0;                    // Not supported
    cAlphaShift     := 0;                    // Not supported
    cAccumBits      := 0;                    // No accumulation buffer
    cAccumRedBits   := 0;                    // Number of red bits in a-buffer
    cAccumGreenBits := 0;                    // Number of green bits in a-buffer
    cAccumBlueBits  := 0;                    // Number of blue bits in a-buffer
    cAccumAlphaBits := 0;                    // Number of alpha bits in a-buffer
    cDepthBits      := PixelDepth;           // Specifies the depth of the depth buffer
    cStencilBits    := 0;                    // Specifies the depth of the stencil buffer
    cAuxBuffers     := 0;                    // Not supported
    iLayerType      := PFD_MAIN_PLANE;       // Ignored
    bReserved       := 0;                    // Number of overlay and underlay planes
    dwLayerMask     := 0;                    // Ignored
    dwVisibleMask   := 0;                    // Transparent color of underlay plane
    dwDamageMask    := 0;                     // Ignored
  end;

  // Attempts to find the pixel format supported by a device context that is the best match to a given pixel format specification.
  PixelFormat := ChoosePixelFormat(h_DC, @pfd);
  if (PixelFormat = 0) then
  begin
    glKillWnd(Fullscreen);
    MessageBox(0, 'Unable to find a suitable pixel format', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;

  // Sets the specified device context's pixel format to the format specified by the PixelFormat.
  if (not SetPixelFormat(h_DC, PixelFormat, @pfd)) then
  begin
    glKillWnd(Fullscreen);
    MessageBox(0, 'Unable to set the pixel format', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;

  // Create a OpenGL rendering context
  h_RC := wglCreateContext(h_DC);
  if (h_RC = 0) then
  begin
    glKillWnd(Fullscreen);
    MessageBox(0, 'Unable to create an OpenGL rendering context', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;

  // Makes the specified OpenGL rendering context the calling thread's current rendering context
  if (not wglMakeCurrent(h_DC, h_RC)) then
  begin
    glKillWnd(Fullscreen);
    MessageBox(0, 'Unable to activate OpenGL rendering context', 'Error', MB_OK or MB_ICONERROR);
    Result := False;
    Exit;
  end;

  // Read And Assign Extentions
  ReadExtensions;
  ReadImplementationProperties;

  if dglCheckExtension('WGL_ARB_create_context') then
  begin

    // Set OpenGL 3.x attribs
    attribs[0]:=WGL_CONTEXT_MAJOR_VERSION_ARB;
    attribs[1]:=OGL3_MAJOR_VERSION_WANTED;;
    attribs[2]:=WGL_CONTEXT_MINOR_VERSION_ARB;
    attribs[3]:=OGL3_MINOR_VERSION_WANTED;
    attribs[4]:=WGL_CONTEXT_FLAGS_ARB;
    attribs[5]:=WGL_CONTEXT_FORWARD_COMPATIBLE_BIT_ARB;
    attribs[6]:=0;

    // Create OpenGL 3.x context using the attribs
    h_RC3 := wglCreateContextAttribsARB(h_DC, 0, @attribs);
    if h_RC3 <> 0 then
    begin
      wglMakeCurrent(0,0); //remove current context
      wglDeleteContext(h_RC); //remove opengl 2.x context
      wglMakeCurrent(h_DC, h_RC3); //make the opengl 3.x context current
      OpenGL3Initialized := True;
    end;
  end else
  begin
    if OGL3_NEEDED then
    begin
      glKillWnd(Fullscreen); //If you do not use OpenGL 3.x specific function you can also continue here ...
      MessageBox(0, 'It is not possible to create an OpenGL 3.x context', 'Error', MB_OK or MB_ICONERROR);
      Result := False;
      Exit;
    end;
  end;

  OpenGLInitialized := True;

  // Used to show OpenGL Version used...
  OGL_TITLE := 'OGL <= 2.x';
  if OpenGL3Initialized then
    OGL_TITLE := 'OGL >= 3.x';

  // Initializes the timer used to calculate the FPS
  SetTimer(h_Wnd, FPS_TIMER, FPS_INTERVAL, nil);

  // Settings to ensure that the window is the topmost window
  ShowWindow(h_Wnd, SW_SHOW);
  SetForegroundWindow(h_Wnd);
  SetFocus(h_Wnd);

  //Init OpenGL
  glInit();

  // Ensure the OpenGL window is resized properly
  glResizeWnd(Width, Height); //Moved after init to enable OpenGL 3.x calls

  Result := True;
end;


{--------------------------------------------------------------------}
{  Main message loop for the application                             }
{--------------------------------------------------------------------}
function WinMain(hInstance : HINST; hPrevInstance : HINST;
                 lpCmdLine : PChar; nCmdShow : Integer) : Integer; stdcall;
var
  msg : TMsg;
  finished : Boolean;
  DemoStart, LastTime : DWord;
begin
  finished := False;

  // Perform application initialization:
  if not glCreateWnd(winwidth, winheight, FALSE, 32) then
  begin
    Result := 0;
    Exit;
  end;

  DemoStart := GetTickCount();            // Get Time when demo started

  // Main message loop:
  while not finished do
  begin
    if (PeekMessage(msg, 0, 0, 0, PM_REMOVE)) then // Check if there is a message for this window
    begin
      if (msg.message = WM_QUIT) then     // If WM_QUIT message received then we are done
        finished := True
      else
      begin                               // Else translate and dispatch the message to this window
  	    TranslateMessage(msg);
        DispatchMessage(msg);
      end;
    end
    else
    begin
      Inc(FPSCount);                      // Increment FPS Counter

      LastTime :=ElapsedTime;
      ElapsedTime :=GetTickCount() - DemoStart;     // Calculate Elapsed Time
      ElapsedTime :=(LastTime + ElapsedTime) DIV 2; // Average it out for smoother movement

      glDraw();                           // Draw the scene
      SwapBuffers(h_DC);                  // Display the scene

      if (keys[VK_ESCAPE]) then           // If user pressed ESC then set finised TRUE
        finished := True
      else
      begin
        ProcessKeys;                      // Check for any other key Pressed
        ProcessMouseMove;                 // Check for mouse move
        ProcessMouseClick;                // Check for Mouse Click
      end;
    end;
  end;
  glKillWnd(FALSE);
  Result := msg.wParam;
end;


{------------------------------------------------------------------}
{  Run OpenGL Application                                          }
{------------------------------------------------------------------}
begin
  WinMain( hInstance, hPrevInst, CmdLine, CmdShow );
end.

